### -*- coding: utf-8 -*- ####################################################
##############################################################################
#
# Copyright (c) 2012 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################


# import standard packages

# import Zope3 interfaces

# import local interfaces

# import Zope3 packages

# import local packages
from ztfy.utils.unicode import translateString


def getContentName(container, base_name, translate=True, max_length=30):
    """Get a real name for a given base name and a container
    
    Target name will be suffixed with an index if base name already exists
    """
    if translate:
        base_name = translateString(base_name, escapeSlashes=True, spaces='-')
    if max_length:
        base_name = base_name[0:max_length]
    if base_name not in container:
        return base_name
    index = 2
    name = '%s-%02d' % (base_name, index)
    while name in container:
        index += 1
        name = '%s-%02d' % (base_name, index)
    return name



### -*- coding: utf-8 -*- ####################################################
##############################################################################
#
# Copyright (c) 2008-2012 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

__docformat__ = "restructuredtext"

# import standard packages

# import Zope3 interfaces
from zope.schema.interfaces import IVocabularyFactory, IChoice

# import local interfaces

# import Zope3 packages
from zope.i18n import translate
from zope.interface import classProvides, implements
from zope.schema import Choice
from zope.schema.vocabulary import SimpleTerm, SimpleVocabulary

# import local packages
from ztfy.utils.request import queryRequest

from ztfy.utils import _


ENCODINGS = {
              'ascii': _('English (ASCII)'),
              'big5': _('Traditional Chinese (big5)'),
              'big5hkscs': _('Traditional Chinese (big5hkscs)'),
              'cp037': _('English (cp037)'),
              'cp424': _('Hebrew (cp424)'),
              'cp437': _('English (cp437)'),
              'cp500': _('Western Europe (cp500)'),
              'cp720': _('Arabic (cp720)'),
              'cp737': _('Greek (cp737)'),
              'cp775': _('Baltic languages (cp775)'),
              'cp850': _('Western Europe (cp850)'),
              'cp852': _('Central and Eastern Europe (cp852)'),
              'cp855': _('Bulgarian, Byelorussian, Macedonian, Russian, Serbian (cp855)'),
              'cp856': _('Hebrew (cp856)'),
              'cp857': _('Turkish (cp857)'),
              'cp858': _('Western Europe (cp858)'),
              'cp860': _('Portuguese (cp860)'),
              'cp861': _('Icelandic (cp861)'),
              'cp862': _('Hebrew (cp862)'),
              'cp863': _('Canadian (cp863)'),
              'cp864': _('Arabic (cp864)'),
              'cp865': _('Danish, Norwegian (cp865)'),
              'cp866': _('Russian (cp866)'),
              'cp869': _('Greek (cp869)'),
              'cp874': _('Thai (cp874)'),
              'cp875': _('Greek (cp875)'),
              'cp932': _('Japanese (cp932)'),
              'cp949': _('Korean (cp949)'),
              'cp950': _('Traditional Chinese (cp950)'),
              'cp1006': _('Urdu (cp1006)'),
              'cp1026': _('Turkish (cp1026)'),
              'cp1140': _('Western Europe (cp1140)'),
              'cp1250': _('Central and Eastern Europe (cp1250)'),
              'cp1251': _('Bulgarian, Byelorussian, Macedonian, Russian, Serbian (cp1251)'),
              'cp1252': _('Western Europe (cp1252)'),
              'cp1253': _('Greek (cp1253)'),
              'cp1254': _('Turkish (cp1254)'),
              'cp1255': _('Hebrew (cp1255)'),
              'cp1256': _('Arabic (cp1256)'),
              'cp1257': _('Baltic languages (cp1257)'),
              'cp1258': _('Vietnamese (cp1258)'),
              'euc_jp': _('Japanese (euc_jp)'),
              'euc_jis_2004': _('Japanese (euc_jis_2004)'),
              'euc_jisx0213': _('Japanese (euc_jisx0213)'),
              'euc_kr': _('Korean (euc_kr)'),
              'gb2312': _('Simplified Chinese (gb2312)'),
              'gbk': _('Unified Chinese (gbk)'),
              'gb18030': _('Unified Chinese (gb18030)'),
              'hz': _('Simplified Chinese (hz)'),
              'iso2022_jp': _('Japanese (iso2022_jp)'),
              'iso2022_jp_1': _('Japanese (iso2022_jp_1)'),
              'iso2022_jp_2': _('Japanese, Korean, Simplified Chinese, Western Europe, Greek (iso2022_jp_2)'),
              'iso2022_jp_2004': _('Japanese (iso2022_jp_2004)'),
              'iso2022_jp_3': _('Japanese (iso2022_jp_3)'),
              'iso2022_jp_ext': _('Japanese (iso2022_jp_ext)'),
              'iso2022_kr': _('Korean (iso2022_kr)'),
              'latin_1': _('West Europe (latin_1)'),
              'iso8859_2': _('Central and Eastern Europe (iso8859_2)'),
              'iso8859_3': _('Esperanto, Maltese (iso8859_3)'),
              'iso8859_4': _('Baltic languages (iso8859_4)'),
              'iso8859_5': _('Bulgarian, Byelorussian, Macedonian, Russian, Serbian (iso8859_5)'),
              'iso8859_6': _('Arabic (iso8859_6)'),
              'iso8859_7': _('Greek (iso8859_7)'),
              'iso8859_8': _('Hebrew (iso8859_8)'),
              'iso8859_9': _('Turkish (iso8859_9)'),
              'iso8859_10': _('Nordic languages (iso8859_10)'),
              'iso8859_13': _('Baltic languages (iso8859_13)'),
              'iso8859_14': _('Celtic languages (iso8859_14)'),
              'iso8859_15': _('Western Europe (iso8859_15)'),
              'iso8859_16': _('South-Eastern Europe (iso8859_16)'),
              'johab': _('Korean (johab)'),
              'koi8_r': _('Russian (koi8_r)'),
              'koi8_u': _('Ukrainian (koi8_u)'),
              'mac_cyrillic': _('Bulgarian, Byelorussian, Macedonian, Russian, Serbian (mac_cyrillic)'),
              'mac_greek': _('Greek (mac_greek)'),
              'mac_iceland': _('Icelandic (mac_iceland)'),
              'mac_latin2': _('Central and Eastern Europe (mac_latin2)'),
              'mac_roman': _('Western Europe (mac_roman)'),
              'mac_turkish': _('Turkish (mac_turkish)'),
              'ptcp154': _('Kazakh (ptcp154)'),
              'shift_jis': _('Japanese (shift_jis)'),
              'shift_jis_2004': _('Japanese (shift_jis_2004)'),
              'shift_jisx0213': _('Japanese (shift_jisx0213)'),
              'utf_32': _('all languages (utf_32)'),
              'utf_32_be': _('all languages (utf_32_be)'),
              'utf_32_le': _('all languages (utf_32_le)'),
              'utf_16': _('all languages (utf_16)'),
              'utf_16_be': _('all languages (BMP only - utf_16_be)'),
              'utf_16_le': _('all languages (BMP only - utf_16_le)'),
              'utf_7': _('all languages (utf_7)'),
              'utf_8': _('all languages (utf_8)'),
              'utf_8_sig': _('all languages (utf_8_sig)'),
             }


class EncodingsVocabulary(SimpleVocabulary):

    classProvides(IVocabularyFactory)

    def __init__(self, terms, *interfaces):
        request = queryRequest()
        terms = [SimpleTerm(unicode(v), title=translate(t, context=request)) for v, t in ENCODINGS.iteritems()]
        terms.sort(key=lambda x: x.title)
        super(EncodingsVocabulary, self).__init__(terms, *interfaces)


class IEncodingField(IChoice):
    """Encoding field interface"""


class EncodingField(Choice):
    """Encoding field"""

    implements(IEncodingField)

    def __init__(self, values=None, vocabulary='ZTFY encodings', source=None, **kw):
        if 'values' in kw:
            del kw['values']
        if 'source' in kw:
            del kw['source']
        kw['vocabulary'] = vocabulary
        super(EncodingField, self).__init__(**kw)



### -*- coding: utf-8 -*- ####################################################
##############################################################################
#
# Copyright (c) 2012 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################


# import standard packages

# import Zope3 interfaces

# import local interfaces

# import Zope3 packages

# import local packages


def unique(seq, idfun=None):
    """Get list unique values, preserving order"""
    if idfun is None:
        def idfun(x): return x
    seen = {}
    result = []
    for item in seq:
        marker = idfun(item)
        if marker in seen:
            continue
        seen[marker] = 1
        result.append(item)
    return result



### -*- coding: utf-8 -*- ####################################################
##############################################################################
#
# Copyright (c) 2008-2012 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

__docformat__ = "restructuredtext"

# import standard packages
from persistent import Persistent
from persistent.interfaces import IPersistent
from transaction.interfaces import ITransactionManager

# import Zope3 interfaces
from ZODB.interfaces import IConnection
from zope.schema.interfaces import IVocabularyFactory

# import local interfaces
from ztfy.utils.interfaces import IZEOConnection

# import Zope3 packages
from ZEO import ClientStorage
from ZODB import DB
from zope.component import adapter
from zope.componentvocabulary.vocabulary import UtilityVocabulary
from zope.container.contained import Contained
from zope.interface import implementer, implements, classProvides
from zope.schema import getFieldNames
from zope.schema.fieldproperty import FieldProperty

# import local packages


class ZEOConnectionInfo(object):
    """ZEO connection info
    
    Provides a context manager which directly returns
    ZEO connection root
    """

    implements(IZEOConnection)

    _storage = None
    _db = None
    _connection = None

    server_name = FieldProperty(IZEOConnection['server_name'])
    server_port = FieldProperty(IZEOConnection['server_port'])
    storage = FieldProperty(IZEOConnection['storage'])
    username = FieldProperty(IZEOConnection['username'])
    password = FieldProperty(IZEOConnection['password'])
    server_realm = FieldProperty(IZEOConnection['server_realm'])
    blob_dir = FieldProperty(IZEOConnection['blob_dir'])
    shared_blob_dir = FieldProperty(IZEOConnection['shared_blob_dir'])

    def getSettings(self):
        result = {}
        for name in getFieldNames(IZEOConnection):
            result[name] = getattr(self, name)
        return result

    def update(self, settings):
        names = getFieldNames(IZEOConnection)
        for k, v in settings.items():
            if k in names:
                setattr(self, k, unicode(v) if isinstance(v, str) else v)

    def getConnection(self, wait=False, get_storage=False):
        """Get a tuple made of storage and DB connection for given settings"""
        storage = ClientStorage.ClientStorage((str(self.server_name), self.server_port),
                                              storage=self.storage,
                                              username=self.username or '',
                                              password=self.password or '',
                                              realm=self.server_realm,
                                              blob_dir=self.blob_dir,
                                              shared_blob_dir=self.shared_blob_dir,
                                              wait=wait)
        db = DB(storage)
        return (storage, db) if get_storage else db

    @property
    def connection(self):
        return self._connection

    # Context management methods
    def __enter__(self):
        self._storage, self._db = self.getConnection(get_storage=True)
        self._connection = self._db.open()
        return self._connection.root()

    def __exit__(self, exc_type, exc_value, traceback):
        if self._connection is not None:
            self._connection.close()
        if self._storage is not None:
            self._storage.close()


class ZEOConnectionUtility(ZEOConnectionInfo, Persistent, Contained):
    """Persistent ZEO connection settings utility"""


class ZEOConnectionVocabulary(UtilityVocabulary):
    """ZEO connections vocabulary"""

    classProvides(IVocabularyFactory)

    interface = IZEOConnection
    nameOnly = True


# IPersistent adapters copied from zc.twist package
# also register this for adapting from IConnection
@adapter(IPersistent)
@implementer(ITransactionManager)
def transactionManager(obj):
    conn = IConnection(obj) # typically this will be
                            # zope.app.keyreference.persistent.connectionOfPersistent
    try:
        return conn.transaction_manager
    except AttributeError:
        return conn._txn_mgr
        # or else we give up; who knows.  transaction_manager is the more
        # recent spelling.



### -*- coding: utf-8 -*- ####################################################
##############################################################################
#
# Copyright (c) 2012 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################


# import standard packages

# import Zope3 interfaces

# import local interfaces

# import Zope3 packages
from zope.i18n import translate

# import local packages
from ztfy.utils.request import queryRequest

from ztfy.utils import _


def getHumanSize(value, request=None):
    """Convert given bytes value in human readable format"""
    if request is None:
        request = queryRequest()
    if request is not None:
        formatter = request.locale.numbers.getFormatter('decimal')
    else:
        formatter = None
    if value < 1024:
        return translate(_("%d bytes"), context=request) % value
    value = value / 1024.0
    if value < 1024:
        if formatter is None:
            return translate(_("%.1f Kb"), context=request) % value
        else:
            return translate(_("%s Kb"), context=request) % formatter.format(value, '0.0')
    value = value / 1024.0
    if value < 1024:
        if formatter is None:
            return translate(_("%.2f Mb"), context=request) % value
        else:
            return translate(_("%s Mb"), context=request) % formatter.format(value, '0.00')
    value = value / 1024.0
    if formatter is None:
        return translate(_("%.3f Gb"), context=request) % value
    else:
        return translate(_("%s Gb"), context=request) % formatter.format(value, '0.000')



### -*- coding: utf-8 -*- ####################################################
##############################################################################
#
# Copyright (c) 2012 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################


# import standard packages
import decimal
import string

# import Zope3 interfaces
from z3c.form.interfaces import IWidget
from zope.schema.interfaces import ITextLine, IDecimal, IList, ITuple

# import local interfaces

# import Zope3 packages
from z3c.form.converter import BaseDataConverter, FormatterValidationError
from zope.component import adapts
from zope.interface import implements
from zope.schema import TextLine, Decimal, List, Tuple
from zope.schema._bootstrapfields import InvalidValue

# import local packages

from ztfy.utils import _


class StringLine(TextLine):
    """String line field"""

    _type = str

    def fromUnicode(self, value):
        return str(value)


#
# Color field
#

class IColorField(ITextLine):
    """Marker interface for color fields"""


class ColorField(TextLine):
    """Color field"""

    implements(IColorField)

    def __init__(self, *args, **kw):
        super(ColorField, self).__init__(max_length=6, *args, **kw)

    def _validate(self, value):
        if len(value) not in (3, 6):
            raise InvalidValue, _("Color length must be 3 or 6 characters")
        for v in value:
            if v not in string.hexdigits:
                raise InvalidValue, _("Color value must contain only valid color codes (numbers or letters between 'A' end 'F')")
        super(ColorField, self)._validate(value)


#
# Pointed decimal field
#

class IDottedDecimalField(IDecimal):
    """Marker interface for dotted decimal fields"""


class DottedDecimalField(Decimal):
    """Dotted decimal field"""

    implements(IDottedDecimalField)


class DottedDecimalDataConverter(BaseDataConverter):
    """Dotted decimal field data converter"""

    adapts(IDottedDecimalField, IWidget)

    errorMessage = _('The entered value is not a valid decimal literal.')

    def __init__(self, field, widget):
        super(DottedDecimalDataConverter, self).__init__(field, widget)

    def toWidgetValue(self, value):
        if not value:
            return self.field.missing_value
        return str(value)

    def toFieldValue(self, value):
        if value is self.field.missing_value:
            return u''
        if not value:
            return None
        try:
            return decimal.Decimal(value)
        except decimal.InvalidOperation:
            raise FormatterValidationError(self.errorMessage, value)


#
# Dates range field
#

class IDatesRangeField(ITuple):
    """Marker interface for dates range fields"""


class DatesRangeField(Tuple):
    """Dates range field"""

    implements(IDatesRangeField)

    def __init__(self, value_type=None, unique=False, **kw):
        super(DatesRangeField, self).__init__(value_type=None, unique=False,
                                              min_length=2, max_length=2, **kw)


#
# TextLine list field
#

class ITextLineListField(IList):
    """Marker interface for textline list field"""


class TextLineListField(List):
    """TextLine list field"""

    implements(ITextLineListField)

    def __init__(self, value_type=None, unique=False, **kw):
        super(TextLineListField, self).__init__(value_type=TextLine(), unique=True, **kw)



### -*- coding: utf-8 -*- ####################################################
##############################################################################
#
# Copyright (c) 2008 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# Python 2.6/2.7 compatibility code copied from EULExistDB
# (https://github.com/emory-libraries/eulexistdb)
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################


# import standard packages
import base64
import cookielib
import httplib
import socket
import urllib2
import xmlrpclib

# import Zope3 interfaces

# import local interfaces

# import Zope3 packages

# import local packages


class TimeoutHTTP(httplib.HTTP):
    def __init__(self, host='', port=None, strict=None, timeout=None):
        if port == 0:
            port = None
        self._setup(self._connection_class(host, port, strict, timeout))

class TimeoutHTTPS(httplib.HTTPS):
    def __init__(self, host='', port=None, strict=None, timeout=None):
        if port == 0:
            port = None
        self._setup(self._connection_class(host, port, strict, timeout))


class XMLRPCCookieAuthTransport(xmlrpclib.Transport):
    """An XML-RPC transport handling authentication via cookies"""

    _http_connection = httplib.HTTPConnection
    _http_connection_compat = TimeoutHTTP

    def __init__(self, user_agent, credentials=(), cookies=None, timeout=socket._GLOBAL_DEFAULT_TIMEOUT, headers=None):
        xmlrpclib.Transport.__init__(self)
        self.user_agent = user_agent
        self.credentials = credentials
        self.cookies = cookies
        self.timeout = timeout
        self.headers = headers
        if self._connection_required_compat():
            self.make_connection = self._make_connection_compat
            self.get_response = self._get_response_compat

    def _connection_required_compat(self):
        # Compatibility code copied from EULExistDB (https://github.com/emory-libraries/eulexistdb)
        # UGLY HACK ALERT. Python 2.7 xmlrpclib caches connection objects in
        # self._connection (and sets self._connection in __init__). Python
        # 2.6 and earlier has no such cache. Thus, if self._connection
        # exists, we're running the newer-style, and if it doesn't then
        # we're running older-style and thus need compatibility mode.
        try:
            self._connection
            return False
        except AttributeError:
            return True

    def make_connection(self, host):
        # This is the make_connection that runs under Python 2.7 and newer.
        # The code is pulled straight from 2.7 xmlrpclib, except replacing
        # HTTPConnection with self._http_connection
        if self._connection and host == self._connection[0]:
            return self._connection[1]
        chost, self._extra_headers, _x509 = self.get_host_info(host)
        self._connection = host, self._http_connection(chost, timeout=self.timeout)
        return self._connection[1]

    def _make_connection_compat(self, host):
        # This method runs as make_connection under Python 2.6 and older.
        # __init__ detects which version we need and pastes this method
        # directly into self.make_connection if necessary.
        host, _extra_headers, _x509 = self.get_host_info(host)
        return self._http_connection_compat(host, timeout=self.timeout)

    # override the send_host hook to also send authentication info
    def send_host(self, connection, host):
        xmlrpclib.Transport.send_host(self, connection, host)
        if (self.cookies is not None) and (len(self.cookies) > 0):
            for cookie in self.cookies:
                connection.putheader('Cookie', '%s=%s' % (cookie.name, cookie.value))
        elif self.credentials:
            auth = 'Basic %s' % base64.encodestring("%s:%s" % self.credentials).strip()
            connection.putheader('Authorization', auth)

    # send custom headers
    def send_headers(self, connection):
        for k, v in (self.headers or {}).iteritems():
            connection.putheader(k, v)

    # dummy request class for extracting cookies
    class CookieRequest(urllib2.Request):
        pass

    # dummy response info headers helper
    class CookieResponseHelper:
        def __init__(self, response):
            self.response = response
        def getheaders(self, header):
            return self.response.msg.getallmatchingheaders(header)

    # dummy response class for extracting cookies
    class CookieResponse:
        def __init__(self, response):
            self.response = response
        def info(self):
            return XMLRPCCookieAuthTransport.CookieResponseHelper(self.response)

    # dummy compat response class for extracting cookies
    class CompatCookieResponse:
        def __init__(self, headers):
            self.headers = headers
        def info(self):
            return self.headers

    def request(self, host, handler, request_body, verbose=False):
        # issue XML-RPC request
        connection = self.make_connection(host)
        self.verbose = verbose
        if verbose:
            connection.set_debuglevel(1)
        self.send_request(connection, handler, request_body)
        self.send_host(connection, host)
        self.send_user_agent(connection)
        self.send_headers(connection)
        self.send_content(connection, request_body)
        # get response
        return self.get_response(connection, host, handler)

    def get_response(self, connection, host, handler):
        response = connection.getresponse()
        # extract cookies from response headers
        if self.cookies is not None:
            crequest = XMLRPCCookieAuthTransport.CookieRequest('http://%s/' % host)
            cresponse = XMLRPCCookieAuthTransport.CookieResponse(response)
            self.cookies.extract_cookies(cresponse, crequest)
        if response.status != 200:
            raise xmlrpclib.ProtocolError(host + handler, response.status, response.reason, response.getheaders())
        return self.parse_response(response)

    def _get_response_compat(self, connection, host, handler):
        errcode, errmsg, headers = connection.getreply()
        # extract cookies from response headers
        if self.cookies is not None:
            crequest = XMLRPCCookieAuthTransport.CookieRequest('http://%s/' % host)
            cresponse = XMLRPCCookieAuthTransport.CompatCookieResponse(headers)
            self.cookies.extract_cookies(cresponse, crequest)
        if errcode != 200:
            raise xmlrpclib.ProtocolError(host + handler, errcode, errmsg, headers)
        try:
            sock = connection._conn.sock
        except AttributeError:
            sock = None
        return self._parse_response(connection.getfile(), sock)


class SecureXMLRPCCookieAuthTransport(XMLRPCCookieAuthTransport):
    """Secure XML-RPC transport"""

    _http_connection = httplib.HTTPSConnection
    _http_connection_compat = TimeoutHTTPS


def getClient(uri, credentials=(), verbose=False, allow_none=0, timeout=socket._GLOBAL_DEFAULT_TIMEOUT, headers=None):
    """Get an XML-RPC client which supports basic authentication"""
    if uri.startswith('https:'):
        transport = SecureXMLRPCCookieAuthTransport('Python XML-RPC Client/0.1 (ZTFY secure transport)', credentials, timeout=timeout, headers=headers)
    else:
        transport = XMLRPCCookieAuthTransport('Python XML-RPC Client/0.1 (ZTFY basic transport)', credentials, timeout=timeout, headers=headers)
    return xmlrpclib.Server(uri, transport=transport, verbose=verbose, allow_none=allow_none)


def getClientWithCookies(uri, credentials=(), verbose=False, allow_none=0, timeout=socket._GLOBAL_DEFAULT_TIMEOUT, headers=None):
    """Get an XML-RPC client which supports authentication through cookies"""
    if uri.startswith('https:'):
        transport = SecureXMLRPCCookieAuthTransport('Python XML-RPC Client/0.1 (ZTFY secure cookie transport)', credentials, cookielib.CookieJar(), timeout, headers)
    else:
        transport = XMLRPCCookieAuthTransport('Python XML-RPC Client/0.1 (ZTFY basic cookie transport)', credentials, cookielib.CookieJar(), timeout, headers)
    return xmlrpclib.Server(uri, transport=transport, verbose=verbose, allow_none=allow_none)
